import subprocess
from abc import ABC, abstractmethod
from concurrent.futures import Executor, ProcessPoolExecutor, as_completed
from multiprocessing import cpu_count
from random import randint
from typing import Iterator, Optional

from diopter.compiler import Language, SourceProgram
from diopter.sanitizer import Sanitizer


class Generator(ABC):
    def __init__(self, sanitizer: Sanitizer):
        self.sanitizer = sanitizer

    @abstractmethod
    def generate_program_impl(self) -> SourceProgram:
        """Concrete subclasses must implement this

        Returns:
            SourceProgram: generated program
        """
        pass

    @abstractmethod
    def filter_program(self, program: SourceProgram) -> bool:
        """Concrete subclasses must implement this to check if the generated
           program should be discared
        Args:
            program (SourceProgram): the program to check
        Returns:
            bool: if the program is good(sanitized)
        """
        pass

    def generate_program(self) -> SourceProgram:
        while True:
            program = self.generate_program_impl()
            if self.filter_program(program):
                return program

    def generate_programs_parallel(
        self, n: int, e: Optional[Executor] = None, jobs: Optional[int] = None
    ) -> Iterator[SourceProgram]:
        """
        Args:
            n (int): how many cases to generate
            e (Optional[Executor]): executor used for running the code generation jobs
            jobs (Optional[int]):
                number of jobs, if None cpu_count() is used (ignored if e is not None)
        Returns:
            Iterator[SourceProgram]: the generated programs
        """

        def make_futures(e: Executor) -> Iterator[SourceProgram]:
            futures = (e.submit(self.generate_program) for _ in range(n))
            for future in as_completed(futures):
                yield future.result()

        if e:
            yield from make_futures(e)
        else:
            with ProcessPoolExecutor(jobs if jobs else cpu_count()) as p:
                yield from make_futures(p)


class CSmithGenerator(Generator):
    default_options_pool = [
        "arrays",
        "bitfields",
        "checksum",
        "comma-operators",
        "compound-assignment",
        "consts",
        "divs",
        "embedded-assigns",
        "jumps",
        "longlong",
        "force-non-uniform-arrays",
        "math64",
        "muls",
        "packed-struct",
        "paranoid",
        "pointers",
        "structs",
        "inline-function",
        "return-structs",
        "arg-structs",
        "dangling-global-pointers",
    ]
    fixed_options = [
        "--no-unions",
        "--safe-math",
        "--no-argc",
        "--no-volatiles",
        "--no-volatile-pointers",
    ]

    def __init__(
        self,
        sanitizer: Sanitizer,
        csmith: Optional[str] = None,
        include_path: Optional[str] = None,
        options_pool: Optional[list[str]] = None,
        minimum_length: int = 10000,
        maximum_length: int = 50000,
    ):
        """
        Args:
            sanitizer (Sanitizer):
                used to sanitize and discard generated code
            csmith (Optional[str]):
                Path to csmith executable, if empty "csmith" will be used
            include_path (Optional[str]):
                csmith include path, if empty "/usr/include/csmith-2.3.0" will be used
            options_pool (Optional[list[str]]):
                csmith options that will be randomly selected,
                if empty default_options_pool will be used
            minimum_length (int):
                The minimum length of a generated test case in characters.
            maximum_length (int):
                The maximum length of a generated test case in characters.
        """
        super().__init__(sanitizer)
        self.minimum_length = minimum_length
        self.maximum_length = maximum_length
        self.csmith = csmith if csmith else "csmith"
        self.options = (
            options_pool if options_pool else CSmithGenerator.default_options_pool
        )
        # XXX: don't hardcode the path here
        self.include_path = (
            include_path if include_path else "/usr/include/csmith-2.3.0"
        )

    def generate_program_impl(self) -> SourceProgram:
        """Generate random code with csmith.

        Returns:
            SourceProgram: csmith generated program.
        """

        cmd = [self.csmith] + CSmithGenerator.fixed_options
        for option in self.options:
            if randint(0, 1):
                cmd.append(f"--{option}")
            else:
                cmd.append(f"--no-{option}")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        assert result.returncode == 0
        return SourceProgram(
            code=result.stdout.decode("utf-8"),
            language=Language.C,
            available_macros=(),
            defined_macros=(),
            include_paths=(),
            system_include_paths=(self.include_path,),
            flags=(),
        )

    def filter_program(self, program: SourceProgram) -> bool:
        if (
            len(program.code) < self.minimum_length
            or len(program.code) > self.maximum_length
        ):
            return False
        return bool(self.sanitizer.sanitize(program))



import inspect
import logging
import os
import pickle
import subprocess
import sys
from abc import ABC, abstractmethod
from multiprocessing import cpu_count
from pathlib import Path
from shutil import which
from sys import stderr
from typing import Optional, TextIO

from diopter.utils import TempDirEnv, run_cmd_to_logfile


class ReductionCallback(ABC):
    @abstractmethod
    def test(self, code: str) -> bool:
        pass


def emit_module_imports(reduction_callback: ReductionCallback) -> str:
    callback_name = type(reduction_callback).__name__
    callback_module_path = inspect.getsourcefile(type(reduction_callback))
    assert callback_module_path
    callback_module = inspect.getmodulename(callback_module_path)
    sys_path_append = "".join(f'\nsys.path.append("{p}")' for p in sys.path)

    return f"""import importlib
import pickle
import sys
from pathlib import Path
sys.path.insert(0, "{str(Path(callback_module_path).parent)}")
{sys_path_append}

from {callback_module} import {callback_name}
"""


def emit_call(reduction_callback: ReductionCallback, code_filename: str) -> str:
    callback_in_hex = pickle.dumps(reduction_callback).hex()
    callback_load = f'callback = pickle.loads(bytes.fromhex("{callback_in_hex}"))'

    call = "exit(not callback.test(code))"
    return f"""with open(\"{code_filename}\", \"r\") as f:
    code = f.read()
{callback_load}
{call}
    """


def make_interestingness_script(
    reduction_callback: ReductionCallback, code_filename: str
) -> str:
    """
    Helper function to create a script useful for use with diopter.Reducer.
    It serializes the reduction callback into a script that checks if the code in
    code_filename is still interesting

    Args:
        reduction_callback:
            callback that will be serialized stored as a runnable script
        code_filename:
            file that the generated script will check everytime it is run

    Returns:
        The python3 script (str)
    """
    prologue = f"#!{sys.executable}"
    return "\n".join(
        (
            prologue,
            emit_module_imports(reduction_callback),
            emit_call(reduction_callback, code_filename),
        )
    )


class Reducer:
    """
    Reducer is a wrapper around CReduce
    """

    def __init__(self, creduce: Optional[str] = None):
        """
        Args:
            creduce: path to the creduce binary, if empty "creduce" will be
            used
        """
        self.creduce = creduce if creduce else "creduce"
        assert which(self.creduce), f"{self.creduce} is not executable"

    def reduce(
        self,
        code: str,
        interestingness_test: ReductionCallback,
        jobs: Optional[int] = None,
        log_file: Optional[TextIO] = None,
        debug: bool = False,
    ) -> Optional[str]:
        """
        Reduce given code

        Args:
            code:
                the code to reduce
            interestingness_test:
                a concrete ReductionCallback that implementes the interestingness
                (can be generated with make_interestingness_check).
            jobs:
                The number of Creduce jobs, if empty cpu_count() will be
            log_file:
                Where to log Creduce's output, if empty stderr will be used
            debug:
                Whether to pass the debug flag to creduce

        Returns:
            Reduced code, if successful.
        """
        creduce_jobs = jobs if jobs else cpu_count()

        interestingness_script = make_interestingness_script(
            interestingness_test, "code.c"
        )

        # creduce likes to kill unfinished processes with SIGKILL
        # so they can't clean up after themselves.
        # Setting a temporary temporary directory for creduce to be able to clean
        # up everything
        with TempDirEnv() as tmpdir:

            code_file = tmpdir / "code.c"
            with open(code_file, "w") as f:
                f.write(code)

            script_path = tmpdir / "check.py"
            with open(script_path, "w") as f:
                print(interestingness_script, file=f)
            os.chmod(script_path, 0o770)
            # run creduce
            creduce_cmd = [
                self.creduce,
                "--n",
                f"{creduce_jobs}",
                str(script_path.name),
                str(code_file.name),
            ]
            if debug:
                creduce_cmd.append("--debug")

            try:
                run_cmd_to_logfile(
                    creduce_cmd,
                    log_file=log_file if log_file else stderr,
                    working_dir=Path(tmpdir),
                    additional_env={"TMPDIR": str(tmpdir.absolute())},
                )
            except subprocess.CalledProcessError as e:
                logging.info(f"Failed to reduce code. Exception: {e}")
                return None

            with open(code_file, "r") as f:
                reduced_code = f.read()

            return reduced_code



""" Preprocessing of csmith files in a compiler portable manner

Example:

csmith_program: SourceProgram = ...
preprocessed_csmith_program = preprocess_csmith_program(csmith_program)
if preprocess_csmith_program:
    # preprocessing was successful
"""

import re
from dataclasses import replace
from typing import Optional

from diopter.compiler import (
    CompilationSetting,
    CompileError,
    CompilerExe,
    OptLevel,
    SourceProgram,
)


def preprocess_lines(lines: list[str]) -> str:
    start_patterns = [
        re.compile(r"^extern.*"),
        re.compile(r"^typedef.*"),
        re.compile(r"^struct.*"),
        # The following patterns are to catch if the last of the previous
        # patterns in the file was tainted and we'd otherwise mark the rest
        # of the file as tainted, as we'll find no end in this case.
        re.compile(r"^static.*"),
        re.compile(r"^void.*"),
    ]
    taint_patterns = [
        re.compile(r".*__access__.*"),  # LLVM doesn't know about this
        re.compile(r".*__malloc__.*"),
        re.compile(
            r".*_[F|f]loat[0-9]{1,3}x{0,1}.*"
        ),  # https://gcc.gnu.org/onlinedocs/gcc/Floating-Types.html#Floating-Types
        re.compile(r".*__asm__.*"),  # CompCert has problems
    ]

    def is_start(line: str) -> bool:
        return any([p_start.match(line) for p_start in start_patterns])

    lines_to_skip: list[int] = []
    for i, line in enumerate(lines):
        for p in taint_patterns:
            if p.match(line):
                # Searching for start of tainted region
                up_i = i
                up_line = lines[up_i]
                while up_i > 0 and not is_start(up_line):
                    up_i -= 1
                    up_line = lines[up_i]

                # Searching for end of tainted region
                down_i = i + 1
                down_line = lines[down_i]
                while down_i < len(lines) and not is_start(down_line):
                    down_i += 1
                    down_line = lines[down_i]

                lines_to_skip.extend(list(range(up_i, down_i)))

    return "\n".join([line for i, line in enumerate(lines) if i not in lines_to_skip])


def preprocess_csmith_program(
    program: SourceProgram, compiler: CompilerExe
) -> Optional[SourceProgram]:
    """Will *try* to preprocess code as if it comes from csmith.

    Args:
        program (SourceProgram):
            program to preprocess
        compiler (CompilerExe):
            the compiler to use for preprocessing

    Returns:
        Optional[SourceProgram]:
            preprocessed program if it was able to preprocess it.
    """

    try:
        result = CompilationSetting(
            compiler=compiler, opt_level=OptLevel.O0
        ).preprocess_program(program)
        return replace(
            program,
            code=preprocess_lines(result.stdout_stderr_output.split("\n")),
            system_include_paths=tuple(
                path
                for path in program.system_include_paths
                if not path.endswith("csmith.h")
            ),
        )
    except CompileError:
        return None



import logging
import math
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from ccbuilder import Commit, CompilerProject, Repo, Revision, find_cached_revisions


class BisectionException(Exception):
    pass


def find_sorted_cached_commits_from_range(
    good_rev: Revision,
    bad_rev: Revision,
    project: CompilerProject,
    repo: Repo,
    cache_prefix: Path,
) -> list[Commit]:
    """Return all cached commits that are in the range R=[good_rev, bad_rev] such that
    R[i+1] is ancestor of R[i].

    Args:
        good_rev (Revision): good_rev
        bad_rev (Revision): bad_rev
        project (CompilerProject): project
        repo (Repo): Repository for `project`
        cache_prefix (Path): Path to the cache.

    Returns:
        list[Commit]:
            All cached commits in [good_rev, bad_rev] ordered via the ancestor relation.
    """
    possible_revs = repo.direct_first_parent_path(good_rev, bad_rev)

    cached_revs = find_cached_revisions(project, cache_prefix)
    cached_revs = [r for r in cached_revs if r in possible_revs]

    # Create enumeration dict to sort cached_revs with
    sort_dict = dict((r, v) for v, r in enumerate(possible_revs))
    cached_revs = sorted(cached_revs, key=lambda x: sort_dict[x])
    return cached_revs


def _get_midpoint_after_failure(
    bad_rev: Revision,
    good_rev: Revision,
    midpoint: Revision,
    failed_to_build_counter: int,
    repo: Repo,
    max_failed_builds: int = 3,
) -> Commit:
    if failed_to_build_counter >= max_failed_builds:
        raise BisectionException(
            "Failed too many times in a row while bisecting. Aborting bisection..."
        )
    if failed_to_build_counter % 2 == 0:
        # Get size of range
        range_size = len(repo.direct_first_parent_path(midpoint, bad_rev))

        # Move 10% towards the last bad
        step = max(int(0.9 * range_size), 1)
        midpoint = repo.rev_to_commit(f"{bad_rev}~{step}")
    else:
        # Symmetric to case above but jumping 10% into
        # the other direction i.e 20% from our position.
        range_size = len(repo.direct_first_parent_path(good_rev, midpoint))
        step = max(int(0.2 * range_size), 1)
        midpoint = repo.rev_to_commit(f"{midpoint}~{step}")
    return midpoint


def _terminate(guaranteed_termination_counter: int) -> bool:
    if guaranteed_termination_counter >= 20:
        logging.warning(
            "Failed too many times in a row while bisecting. Aborting bisection..."
        )
        return True
    return False


class BisectionCallback(ABC):
    @abstractmethod
    def check(self, commit: Commit) -> Optional[bool]:
        pass


class Bisector:
    def __init__(self, build_cache_prefix: Path):
        self.build_cache_prefix = build_cache_prefix
        return

    def bisect(
        self,
        test: BisectionCallback,
        bad_rev: Revision,
        good_rev: Revision,
        project: CompilerProject,
        repo: Repo,
    ) -> Optional[Commit]:
        """Bisect between `bad_rev` and `good_rev` based on the provided `test` callback
        `test(bad_rev)` must be True.
        `test(good_rev)` must be False.

        Args:
            test (BisectionCallback):
                Interestingness test.
            bad_rev (Revision):
                Revision that shows the initial unwanted behaviour, defined in `test`.
            good_rev (Revision):
                Revision that shows the initial correct behaviour.
            project (CompilerProject):
                Which project on bisect in.
            repo (Repo):
                Repository for the `project`.

        Returns:
            Optional[Commit]: The bisection commit that introduced the behaviour
        """

        # sanitize revs
        good_commit = repo.rev_to_commit(good_rev)
        bad_commit = repo.rev_to_commit(bad_rev)

        # Make sure there is a direct path between the two
        # commits. If none is found or the good commit would
        # 'land on top', we abort.
        if not repo.is_ancestor(good_commit, bad_commit):
            bca = repo.get_best_common_ancestor(good_commit, bad_commit)
            test_res: Optional[bool] = test.check(bca)
            match test_res:
                case False:
                    good_commit = bca
                case True:
                    logging.info("Best common ancestor is interesting. Can't bisect.")
                    return None
                case None:
                    logging.info("Test for best common ancestor failed. Can't bisect.")
                    return None

        try:
            good_commit, bad_commit = self._in_cache_path_bisection(
                test=test,
                bad_commit=bad_commit,
                good_commit=good_commit,
                project=project,
                repo=repo,
            )
            logging.info(f"{good_commit=} {bad_commit=}")
            if bisection_commit := self._normal_path_bisection(
                test=test,
                bad_commit=bad_commit,
                good_commit=good_commit,
                project=project,
                repo=repo,
            ):
                # Check if the result is correct
                pre_bisection_commit = repo.rev_to_commit(f"{bisection_commit}~")
                bisection_res = test.check(bisection_commit)
                pre_bisection_res = test.check(pre_bisection_commit)
                if bisection_res and not pre_bisection_res:
                    return bisection_commit
                logging.warning("Bisection check failed!")
                return None
            return None
        except BisectionException as e:
            logging.warning(e)
            return None

    def _in_cache_path_bisection(
        self,
        test: BisectionCallback,
        bad_commit: Commit,
        good_commit: Commit,
        project: CompilerProject,
        repo: Repo,
    ) -> tuple[Commit, Commit]:

        cached_commits = find_sorted_cached_commits_from_range(
            good_commit, bad_commit, project, repo, self.build_cache_prefix
        )

        logging.info("Bisecting in cache...")
        midpoint = ""
        old_midpoint = ""
        while True:
            logging.info(
                f"{len(cached_commits): 4}, bad: {bad_commit}, good: {good_commit}"
            )
            if len(cached_commits) == 0:
                break
            midpoint_idx = len(cached_commits) // 2
            old_midpoint = midpoint
            midpoint = cached_commits[midpoint_idx]
            if old_midpoint == midpoint:
                break

            test_res: Optional[bool] = test.check(midpoint)

            match test_res:
                case True:
                    bad_commit = midpoint
                    cached_commits = cached_commits[midpoint_idx + 1 :]
                case False:
                    good_commit = midpoint
                    cached_commits = cached_commits[:midpoint_idx]
                case None:
                    # Test failed.
                    cached_commits.remove(midpoint)

        return good_commit, bad_commit

    def _normal_path_bisection(
        self,
        test: BisectionCallback,
        bad_commit: Commit,
        good_commit: Commit,
        project: CompilerProject,
        repo: Repo,
    ) -> Optional[Commit]:

        logging.info("Starting normal bisection...")
        len_region = len(repo.direct_first_parent_path(good_commit, bad_commit))
        logging.info(f"Bisecting for approx. {math.ceil(math.log2(len_region))} steps")
        midpoint = ""
        old_midpoint = ""
        test_failed = False
        failed_to_build_counter = 0

        guaranteed_termination_counter = 0
        while True:
            if not test_failed:
                old_midpoint = midpoint
                midpoint = repo.next_bisection_commit(good_commit, bad_commit)
                failed_to_build_counter = 0
                guaranteed_termination_counter = 0
                if midpoint == "" or midpoint == old_midpoint:
                    break
            else:
                midpoint = _get_midpoint_after_failure(
                    bad_commit, good_commit, midpoint, failed_to_build_counter, repo
                )
                if _terminate(guaranteed_termination_counter):
                    return None
                guaranteed_termination_counter += 1
                failed_to_build_counter += 1
                test_failed = False

            logging.info(f"Midpoint: {midpoint}")

            match test.check(midpoint):
                case True:
                    bad_commit = midpoint
                case False:
                    good_commit = midpoint
                case None:
                    test_failed = True

        logging.info(f"Bisection result: {bad_commit}")
        return bad_commit
